Package: MetaboAnnotation
Title: Utilities for Annotation of Metabolomics Data
Version: 1.14.0
Description: 
    High level functions to assist in annotation of (metabolomics) data sets.
    These include functions to perform simple tentative annotations based on
    mass matching but also functions to consider m/z and retention times for
    annotation of LC-MS features given that respective reference values are
    available. In addition, the function provides high-level functions to
    simplify matching of LC-MS/MS spectra against spectral libraries and objects
    and functionality to represent and manage such matched data.
Authors@R: 
    c(person(given = "Michael", family = "Witting",
             role = c("aut"),
             email = "michael.witting@helmholtz-muenchen.de",
             comment = c(ORCID = "0000-0002-1462-4426")),
      person(given = "Johannes", family = "Rainer",
             email = "Johannes.Rainer@eurac.edu",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-6977-7147")),
      person(given = "Andrea", family = "Vicini",
             email = "andrea.vicini@eurac.edu",
             role = c("aut"),
             comment = c(ORCID = "0000-0001-9438-6909")),
      person(given = "Carolin", family = "Huber",
             email = "carolin-elisabeth.huber@ufz.de",
             role = c("aut"),
             comment = c(ORCID = "0000-0002-9355-8948")),
      person(given = "Philippine", family = "Louail",
             email = "philippine.louail@eurac.edu",
             role = c("aut"),
             comment = c(ORCID = "0009-0007-5429-6846")),
      person(given = "Nir", family = "Shachaf",
             email = "nir.shachaf@weizmann.ac.il",
	     role = "ctb"))
Depends: R (>= 4.0.0)
Imports: BiocGenerics, MsCoreUtils, MetaboCoreUtils, ProtGenerics,
        methods, S4Vectors, Spectra (>= 1.17.6), BiocParallel,
        SummarizedExperiment, QFeatures, AnnotationHub, graphics,
        CompoundDb
Suggests: testthat, knitr, msdata, BiocStyle, rmarkdown, plotly, shiny,
        shinyjs, msentropy, DT, microbenchmark, mzR
Enhances: RMariaDB, RSQLite
License: Artistic-2.0
VignetteBuilder: knitr
BugReports: https://github.com/RforMassSpectrometry/MetaboAnnotation/issues
URL: https://github.com/RforMassSpectrometry/MetaboAnnotation
biocViews: Infrastructure, Metabolomics, MassSpectrometry
Roxygen: list(markdown=TRUE)
RoxygenNote: 7.3.3
Encoding: UTF-8
Collate: 'AllClassUnions.R' 'AllGenerics.R' 'CompAnnotationSource.R'
        'CompDbSource.R' 'Matched.R' 'MatchedSpectra.R'
        'group_standards.R' 'hidden-aliases.R' 'matchFormula.R'
        'matchSpectra.R' 'matchValues.R' 'validateMatchedSpectra.R'
git_url: https://git.bioconductor.org/packages/MetaboAnnotation
git_branch: RELEASE_3_22
git_last_commit: 790d88f
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 04:56:24 UTC; biocbuild
Author: Michael Witting [aut] (ORCID: <https://orcid.org/0000-0002-1462-4426>),
  Johannes Rainer [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6977-7147>),
  Andrea Vicini [aut] (ORCID: <https://orcid.org/0000-0001-9438-6909>),
  Carolin Huber [aut] (ORCID: <https://orcid.org/0000-0002-9355-8948>),
  Philippine Louail [aut] (ORCID:
    <https://orcid.org/0009-0007-5429-6846>),
  Nir Shachaf [ctb]
Maintainer: Johannes Rainer <Johannes.Rainer@eurac.edu>
