# Generated by roxygen2: do not edit by hand

export(CompDbSource)
export(CompareSpectraParam)
export(Mass2MzParam)
export(Mass2MzRtParam)
export(MassBankSource)
export(MatchForwardReverseParam)
export(Matched)
export(MatchedSpectra)
export(Mz2MassParam)
export(Mz2MassRtParam)
export(MzParam)
export(MzRtParam)
export(ScoreThresholdParam)
export(SelectMatchesParam)
export(SingleMatchParam)
export(TopRankedMatchesParam)
export(ValueParam)
export(createStandardMixes)
export(matchFormula)
export(matchMz)
export(matchSpectra)
export(matchValues)
export(matches)
export(pruneTarget)
export(queryIndex)
export(scoreVariables)
export(target)
export(targetIndex)
export(validateMatchedSpectra)
export(whichQuery)
export(whichTarget)
exportClasses(CompAnnotationSource)
exportClasses(CompDbSource)
exportClasses(CompareSpectraParam)
exportClasses(Matched)
exportClasses(MatchedSpectra)
exportMethods("$")
exportMethods("[")
exportMethods(addMatches)
exportMethods(addProcessing)
exportMethods(colnames)
exportMethods(endoapply)
exportMethods(filterMatches)
exportMethods(lapply)
exportMethods(length)
exportMethods(matchSpectra)
exportMethods(matchedData)
exportMethods(metadata)
exportMethods(plotSpectraMirror)
exportMethods(query)
exportMethods(queryVariables)
exportMethods(setBackend)
exportMethods(show)
exportMethods(spectraVariables)
exportMethods(targetVariables)
importClassesFrom(CompoundDb,CompDb)
importClassesFrom(ProtGenerics,Param)
importClassesFrom(QFeatures,QFeatures)
importClassesFrom(S4Vectors,DataFrame)
importClassesFrom(Spectra,MsBackendDataFrame)
importClassesFrom(Spectra,Spectra)
importClassesFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(BiocGenerics,colnames)
importFrom(BiocGenerics,lapply)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpmapply)
importFrom(BiocParallel,bpparam)
importFrom(BiocParallel,register)
importFrom(CompoundDb,CompDb)
importFrom(MetaboCoreUtils,adductNames)
importFrom(MetaboCoreUtils,mass2mz)
importFrom(MetaboCoreUtils,mz2mass)
importFrom(MetaboCoreUtils,standardizeFormula)
importFrom(MsCoreUtils,between)
importFrom(MsCoreUtils,common)
importFrom(MsCoreUtils,ndotproduct)
importFrom(MsCoreUtils,ppm)
importFrom(MsCoreUtils,rbindFill)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,endoapply)
importFrom(Spectra,MsBackendDataFrame)
importFrom(Spectra,Spectra)
importFrom(Spectra,joinPeaks)
importFrom(Spectra,scalePeaks)
importFrom(grDevices,n2mfrow)
importFrom(graphics,par)
importFrom(methods,"slot<-")
importFrom(methods,.hasSlot)
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,slotNames)
importFrom(methods,validObject)
importMethodsFrom(AnnotationHub,query)
importMethodsFrom(BiocGenerics,colnames)
importMethodsFrom(ProtGenerics,as.list)
importMethodsFrom(ProtGenerics,setBackend)
importMethodsFrom(ProtGenerics,spectraVariables)
importMethodsFrom(S4Vectors,"$")
importMethodsFrom(S4Vectors,cbind)
importMethodsFrom(S4Vectors,metadata)
importMethodsFrom(Spectra,"spectraNames<-")
importMethodsFrom(Spectra,addProcessing)
importMethodsFrom(Spectra,backendBpparam)
importMethodsFrom(Spectra,compareSpectra)
importMethodsFrom(Spectra,containsMz)
importMethodsFrom(Spectra,filterPrecursorMzRange)
importMethodsFrom(Spectra,filterRt)
importMethodsFrom(Spectra,peaksData)
importMethodsFrom(Spectra,peaksVariables)
importMethodsFrom(Spectra,plotSpectraMirror)
importMethodsFrom(Spectra,precursorMz)
importMethodsFrom(Spectra,rtime)
importMethodsFrom(Spectra,spectraData)
importMethodsFrom(Spectra,spectraNames)
importMethodsFrom(SummarizedExperiment,rowData)
importMethodsFrom(methods,show)
